#include "limits.h"

typedef unsigned int size_t;
/*@ requires \valid(((char *)dest)+(0..length-1));
requires \valid_read(((char *)src)+(0..length-1));
 requires \separated(((char *)dest)+(0..length -1),((char*)src)+(0..length-1));
assigns ((char *)dest)[0..length-1];
ensures \forall integer i; 0<=i<length ==> ((char*)dest)[i] == ((char*)src)[i];
ensures \result == dest;
*/
void* memcpy(void* dest, void* src, size_t length) {
  /*@loop invariant 0<= i <= length;
    loop invariant \forall integer k; 0<=k<i ==> ((char *)dest)[k] == ((char*)src)[k];
    loop assigns i,((char*)dest)[0..length-1];
    loop variant length - i;
   */ 
 for (size_t i = 0; i<length; i++) {
    ((char*)dest)[i] = ((char*)src)[i];
  }
  return dest;
}

/*@ axiomatic Strlen {
logic integer strlen{L}(char *s) reads s[0..];
axiom strlen{L}: \forall char*s; strlen(s)>= 0 ==> s[strlen(s)] == 0;
axiom strlen_first{L}: \forall char*s, integer i; strlen(s)>=0 ==> 0<=i<strlen(s) ==> s[i] !=0;
axiom strlen_invalid{L}:
\forall char* s; strlen{L}(s) < 0 ==> 
  (\forall integer i; 0<=i ==> s[i] !=0);
}
*/

/*@ predicate valid_string{L}(char *s) = 
  0<= strlen(s) <= UINT_MAX &&
  \valid(s+(0..strlen(s)));
 */

/*@ requires valid_string(s);
  assigns \nothing;
  ensures \result == strlen(s);
*/
size_t strlen(char* s) {
  unsigned int i = 0;
  /*@ loop invariant 0<= i <= strlen(s);
    loop assigns i;
    loop variant strlen(s) - i;
   */
  while (s[i]) 
    i++;
  return i;
}

/* @ lemma has_zero_is_valid{L}: 
  \forall char* s, integer l; 
  0<=l<=UINT_MAX && \valid(s+(0 .. l)) && s[l] == 0 ==>
  valid_string(s);
*/

/*@requires valid_string(s2); 
  requires \valid(s1+(0..strlen(s2)));
requires \separated(s1+(0..strlen(s2)),s2+(0..strlen(s2)));  
assigns s1[0..strlen{Pre}(s2)];
ensures \forall integer i; 0<= i <= strlen{Pre}(s2) ==> s1[i] == s2[i];
ensures valid_string(s1);
ensures \result == s1;
//ensures strlen{Pre}(s2) == strlen(s2);
//ensures strlen(s1) == strlen{Pre}(s2);
 */
char* strcpy(char *s1, const char* s2) {
  size_t i=0;
  /*@ loop invariant 0<=i<=strlen{Pre}(s2);
    loop invariant \forall integer k; 0<=k<i ==> s1[k] == s2[k];
    loop invariant \forall integer k; 0<=k<i ==> s1[k] !=0;
    loop assigns i,s1[0..strlen{Pre}(s2)];
    loop variant strlen{Pre}(s2) - i;
   */
  while (s1[i]=s2[i])
    i++;
  //@ assert s1[i] == 0; 
  return s1;
}

/*@ lemma strlen_def{L}: 
  \forall char *s, integer l; 
  (\forall integer i; 0<= i< l ==> s[i] !=0) && s[l] == 0 ==> strlen(s) == l;
*/

/*@ lemma strlen_long{L}:
  \forall char*s, integer l;
  strlen(s) > l ==> \forall integer i; 0<=i<=l ==> s[i] != 0;
*/

/*@ 
  requires valid_string(s2);
  requires \valid(s1+(0 .. n-1));
  requires \separated(s1+(0 .. n-1), s2+(0 .. strlen(s2)));
  assigns s1[0 .. n-1];
  ensures \result == s1;
  behavior short_string:
  assumes strlen(s2) < n;
  ensures \forall integer i; 0<= i < strlen(s2) ==> s1[i] == s2[i];
  ensures \forall integer i; strlen(s2) <= i < n ==> s1[i] == 0;
  ensures valid_string(s1);
  behavior long_string:
  assumes strlen(s2) >= n;
  ensures \forall integer i; 0<=i<n ==> s1[i] == s2[i];
  
  complete behaviors;
  disjoint behaviors;
 */
char *strncpy(char* s1, const char *s2, size_t n) {
  size_t i = 0;
  /*@
    loop invariant 0<=i<=n;
    loop invariant i<= strlen{Pre}(s2);
    loop assigns i, s1[0 .. n-1]; 
    loop invariant \forall integer j; 0<= j < i ==> s1[j] == s2[j];
    loop invariant \forall integer j; 0<= j < i ==> s2[j] !=0;
    loop variant n - i;
  */
  while (i < n && s2[i] !=0) {
    s1[i] = s2[i];
    i++;
  }
  /*@ for short_string: assert i == strlen{Pre}(s2); */
  /*@ for long_string: assert i == n; */
  //@ assert i == strlen{Pre}(s2) || i == n;
  /*@ loop invariant i<=n;
    loop invariant strlen{Pre}(s2)<=i || i == n;
    loop invariant \forall integer j; 0<= j < strlen{Pre}(s2) && j < n ==> s1[j] == s2[j];
    loop assigns i, s1[0 .. n-1];
    for short_string: loop invariant strlen{Pre}(s2)<=i;
    for short_string: loop invariant \forall integer j; strlen{Pre}(s2)<=j<i ==> s1[j] == 0;
    for long_string: loop invariant i == n;
    loop variant n - i;
   */
  while (i<n) {
    //@ assert i >= strlen{Pre}(s2);
    s1[i] = 0;
    i++;
  }
  return s1;
}

/*
Local Variables:
compile-command: "frama-c-gui -pp-annot string_proved.c -wp -wp-rte -wp-proof why3:Z3,alt-ergo -wp-split -wp-model Typed+ref"
End:
*/
